Falttor = {};
local Falttor_mt = Class(Falttor);
function Falttor.onCreate(id)
	g_currentMission:addUpdateable(Falttor:new(id));
end;
function Falttor:new(id, customMt)
	local instance = {};
	if customMt ~= nil then
		setmetatable(instance, customMt);
	else
		setmetatable(instance, Falttor_mt);
	end;

	instance.triggerId = id;

	instance.manualOpen = Utils.getNoNil(getUserAttribute(id, "manualOpen"), false);
	instance.canBeOpen = 1;
	instance.oldState = instance.canBeOpen;
	instance.playerInTrigger = false;
	if instance.manualOpen == true then
		instance.canBeOpen = 0;
		if getCollisionMask(id) ~= 15736832 then
			setCollisionMask(id,15736832);
		end;
	end;

	addTrigger(id, "triggerCallback", instance);

	local colliId = getUserAttribute(id, "collisionBoxIndex");
	if colliId ~= nil then
		instance.colliIndex = Utils.indexToObject(id, colliId);
		instance.colliMask = getCollisionMask(instance.colliIndex);
	else
		print("Error: Falttor collisionbox index not correct or missing, trigger deleted");
		self:delete();
		return
	end;
	instance.speedScale = Utils.getNoNil(getUserAttribute(id, "speedScale"),10);
	instance.speedScale = instance.speedScale / 20;

	local rootNode = getChild(id, "Sektion_1")
	instance.animCharSet = 0;
	if rootNode ~= nil then
		instance.animCharSet = getAnimCharacterSet(rootNode);
		if instance.animCharSet ~= 0 then
			local clipSource = getUserAttribute(id, "animationClip");
			if clipSource ~= nil then
				instance.clip = getAnimClipIndex(instance.animCharSet, clipSource);
				if instance.clip ~= nil then
					if instance.clip >= 0 then
						assignAnimTrackClip(instance.animCharSet, 0, instance.clip);
						setAnimTrackLoopState(instance.animCharSet, 0, false);
						setAnimTrackSpeedScale(instance.animCharSet, 0, instance.speedScale);
						instance.animDuration = getAnimClipDuration(instance.animCharSet, instance.clip);
					else
						print("Error: Falttor clip for child Sektion_1 are not usable, trigger deleted");
						self:delete();
					end;
				else
					print("Error: Falttor no clip named "..tostring(clipSource).." can be found in animation for child Sektion_1, trigger deleted");
					self:delete();
				end;
			else
				print("Error: Falttor no clip name assigned for child Sektion_1, trigger deleted");
				self:delete();
			end;
		else
			print("Error: Falttor no animation assigned for child Sektion_1, trigger deleted");
			self:delete();
		end;
	else
		print("Error: Falttor no child Sektion_1 are found in assigned gate, trigger deleted");
		self:delete();
	end;

	instance.trafficAllowed = Utils.getNoNil(getUserAttribute(id, "trafficAllowed"), false);

	instance.trackTime = 0;
	instance.lastTrackTime = 0;
	instance.autoCloseTimer = 0;

	local audio = getUserAttribute(id, "audioIndex");
	if audio ~= nil then
		instance.audio = Utils.indexToObject(id, audio);
	else
		print("Error: Falttor Audio index not correct or missing, Trigger deleted");
		self:delete();
	end;
	if instance.audio ~= nil then
		setVisibility(instance.audio, false);
	end;

	instance.count = 0;
	instance.ccount = 0;
	instance.countBackup = 0;

	instance.setOpenState = Falttor.setOpenState;

	instance.saveName = getUserAttribute(id, "triggerName");
	if instance.saveName == nil or instance.saveName == "" then
		local x,y,z = getWorldTranslation(id);
		x = math.floor(x * 100);
		y = math.floor(y * 100);
		z = math.floor(z * 100);
		instance.saveName = convertNumberToAbc(x..y..z);
	end;
	local x, y, z = getWorldTranslation(id);
	if x < 0 then
		x = 0 - x;
	end;
	if y < 0 then
		y = 0- y;
	end;
	if z < 0 then
		z = 0 - z;
	end;
	local sId = math.floor(x * 100) + math.floor(y * 100) + math.floor(z * 100);
	sId = math.floor(sId * 3.5);
	instance.mpName = sId;
	instance.tablePos = -1;
	instance.triggerIsAdded = false;
	instance.isEnabled = true;

	return instance;
end;
function convertNumberToAbc(str)
	replacer = {};
	replacer[1] = "x";
	replacer[2] = "h";
	replacer[3] = "w";
	replacer[4] = "p";
	replacer[5] = "a";
	replacer[6] = "k";
	replacer[7] = "z";
	replacer[8] = "r";
	replacer[9] = "v";
	replacer[10] = "m";
	str = tostring(str);
	for i=1, 10 do
		str = string.gsub(str, tostring(i-1),replacer[i]);
	end;
	str = string.gsub(str, "-","c");
	return str;
end;
function Falttor:delete()
	removeTrigger(self.triggerId);
end;
function Falttor:update(dt)
	-- check if gate in global var, otherwise add it to
	if self.triggerIsAdded == false then
		if g_currentMission.falttorTriggers ~= nil then
			table.insert(g_currentMission.falttorTriggers, self);
			self.tablePos = table.getn(g_currentMission.falttorTriggers);
			self.triggerIsAdded = true;
		else
			g_currentMission.falttorTriggers = {};
			table.insert(g_currentMission.falttorTriggers, self);
			self.tablePos = table.getn(g_currentMission.falttorTriggers);
			self.triggerIsAdded = true;
		end;
	end;

	-- check if manual open true (state 0 close, state 2 open), otherwise set to automatic (state 1)
	if self.manualOpen == true then
		self.closeTime = 0;
		if self.playerInTrigger == true and g_currentMission.player.canDrawOpen then
			if g_gui.currentGui == nil or g_gui.currentGuiName ~= "ChatDialog" then
				if InputBinding.hasEvent(InputBinding.OPEN_GATE) then
					if self.canBeOpen == 0 then
						self.canBeOpen = 2;
					else
						self.canBeOpen = 0;
					end;
					FalttorTriggerManager:setOpenState(self.tablePos, self.mpName, self.canBeOpen);
				end;
			end;
		end;
	elseif self.canBeOpen ~= 1 then
		self.canBeOpen = 1;
	end;

	-- correct tracktime if necessary
	if self.trackTime < 1 then
		self.trackTime = 0;
	end;
	if self.trackTime > self.animDuration then
		self.trackTime = self.animDuration;
	end;

	-- now check if open state to open or autoopen and start animation if trigger count > 0
	if self.canBeOpen > 0 then
		if self.count > 0 or self.canBeOpen > 1 then
			if self.trackTime < self.animDuration then
				self.trackTime = self.trackTime + 10 * self.speedScale;
			end;
			if self.trackTime >= (self.animDuration / 2) then
				setCollisionMask(self.colliIndex,0);
			end;
		else
			if self.trackTime > 0 then
				self.trackTime = self.trackTime - 10 * self.speedScale;
				if self.trackTime <= (self.animDuration / 2) then
					setCollisionMask(self.colliIndex,self.colliMask);
				end;
			end;
		end;
	else
		if self.trackTime > 0 then
			self.trackTime = self.trackTime - 10 * self.speedScale;
			if self.trackTime <= (self.animDuration / 2) then
				setCollisionMask(self.colliIndex,self.colliMask);
			end;
		end;
	end;

	-- set audio visibility and play animation
	if self.lastTrackTime ~= self.trackTime then
		if self.audio ~= nil then
			if not getVisibility(self.audio) then
				setVisibility(self.audio, true);
			end;
		end;
		enableAnimTrack(self.animCharSet, self.clip);
		setAnimTrackTime(self.animCharSet, self.clip, self.trackTime, true);
		disableAnimTrack(self.animCharSet, self.clip);
		self.lastTrackTime = self.trackTime;
	else
		if self.audio ~= nil then
			if getVisibility(self.audio) then
				setVisibility(self.audio, false);
			end;
		end;
	end;
end;
function Falttor:triggerCallback(triggerId, otherId, onEnter, onLeave, onStay)
	if g_currentMission ~= nil then
		if g_currentMission.player ~= nil then
			if g_currentMission.player.rootNode == otherId then
				if onEnter and self.isEnabled then
					if self.manualOpen == true then
						self.playerInTrigger = true;
						g_currentMission.player.canDrawOpen = true;
					end;
					self.count = self.count + self.countBackup;
					self.countBackup = 0;
				elseif onLeave then
					g_currentMission.player.canDrawOpen = false;
					self.playerInTrigger = false;
				end;
			end;
		end;

		local vehicle = g_currentMission.nodeToVehicle[otherId];
		if vehicle ~= nil then
			local canBe = true;
			if SpecializationUtil.hasSpecialization(TrafficVehicle, vehicle.specializations) or SpecializationUtil.hasSpecialization(PathVehicle, vehicle.specializations) then
				if not self.trafficAllowed then
					canBe = false;
				end;
			end;
			if canBe then
				if vehicle.isControlled ~= nil then
					if onEnter and self.isEnabled then
						self.count = self.count + 1;
						self.ccount = self.ccount + 1;
						if self.manualOpen == true then
							self.playerInTrigger = true;
							vehicle.canDrawOpen = true;
						end;
						self.count = self.count + self.countBackup;
						self.countBackup = 0;
					elseif onLeave then
						self.count = self.count - 1;
						if self.countBackup > 0 then
							self.countBackup = self.countBackup - 1;
						end;
						self.ccount = self.ccount - 1;
						if self.ccount <= 0 then
							vehicle.canDrawOpen = false;
							self.playerInTrigger = false;
							self.ccount = 0;
						end;
						if self.count <= 0 then
							vehicle.canDrawOpen = false;
							self.playerInTrigger = false;
							self.count = 0;
						end;
					end;
				else
					if onEnter and self.isEnabled then
						self.count = self.count + 1;
						self.count = self.count + self.countBackup;
						self.countBackup = 0;
					elseif onLeave then
						self.count = self.count - 1;
						if self.countBackup > 0 then
							self.countBackup = self.countBackup - 1;
						end;
						if self.count <= 0 then
							self.count = 0;
							vehicle.canDrawOpen = false;
							self.playerInTrigger = false;
						end;
					end;
				end;
			end;
		end;
	end;
end;
function Falttor:setOpenState(state)
	if state ~= nil then
		self.canBeOpen = state;
	end;
end;
g_onCreateUtil.addOnCreateFunction("FalttorOnCreate", Falttor.onCreate);

-----------------------------------------------------------------------------------

FalttorTriggerManager = {};
FalttorTriggerManager.savegamePath = getUserProfileAppPath() .. "savegame" .. g_careerScreen.selectedIndex;
FalttorTriggerManager.savegameFilename = FalttorTriggerManager.savegamePath .. "/Falttor_states.xml";
FalttorTrigger = {};
FalttorTrigger.modDir = g_currentModDirectory;
FalttorTriggerManagerRandomEvent = {};
FalttorTriggerManagerRandomEvent_mt = Class(FalttorTriggerManagerRandomEvent, Event);
FalttorTriggerManagerEvent = {};
FalttorTriggerManagerEvent_mt = Class(FalttorTriggerManagerEvent, Event);
FalttorTriggerRequestEvent = {};
FalttorTriggerRequestEvent_mt = Class(FalttorTriggerRequestEvent, Event);
InitEventClass(FalttorTriggerManagerRandomEvent, "FalttorTriggerManagerRandomEvent");
InitEventClass(FalttorTriggerRequestEvent, "FalttorTriggerRequestEvent");
InitEventClass(FalttorTriggerManagerEvent, "FalttorTriggerManagerEvent");

local CareerScreenSaveSelectedGame = CareerScreen.saveSelectedGame;
CareerScreen.saveSelectedGame = function(self)
	CareerScreenSaveSelectedGame(self);
	FalttorTriggerManager:saveOpenStates();
end;

function FalttorTrigger:load(xmlFile)
	self.canDrawOpen = false;
end;
function FalttorTriggerManager:loadMap(name)
	self.OTMmapName = name;
	self.filePath = nil;
	self.triggersLoaded = false;
	if self.baseDirectory ~= nil then
		self.filePath = self.baseDirectory;
	elseif g_currentModDirectory ~= nil then
		self.filePath = g_currentModDirectory;
	elseif g_currentMission.missionInfo.baseDirectory ~= nil then
		self.filePath = g_currentMission.missionInfo.baseDirectory;
	elseif g_currentMission.baseDirectory ~= nil then
		self.filePath = g_currentMission.baseDirectory;
	end;
	
	if g_currentMission ~= nil then
		if g_currentMission.falttorTriggers == nil then
			g_currentMission.falttorTriggers = {};
		end;
	end;
	g_currentMission.checkOpenStates = 0;
	self.oldNumTriggers = 0;
	self.syncGates = 100;
end;
function FalttorTriggerManager:deleteMap()
end;
function FalttorTriggerManager:update(dt)
	if (not self.triggersLoaded) then
		if g_currentMission.falttorTriggers ~= nil and table.getn(g_currentMission.falttorTriggers) > 0 then
			if table.getn(g_currentMission.falttorTriggers) > self.oldNumTriggers then
				self.oldNumTriggers = table.getn(g_currentMission.falttorTriggers);
				self:loadOpenStates();
			end;
		end;
	else
		if g_currentMission.missionDynamicInfo.isMultiplayer then
			if not g_server then
				if self.syncGates == 0 then
					if g_currentMission.falttorTriggers ~= nil and table.getn(g_currentMission.falttorTriggers) > 0 then
						for i=1, table.getn(g_currentMission.falttorTriggers) do
							if g_currentMission.falttorTriggers[i] ~= nil then
								if g_currentMission.falttorTriggers[i].manualOpen then
									g_client:getServerConnection():sendEvent(FalttorTriggerRequestEvent:new(g_currentMission.falttorTriggers[i].mpName));
								end;
							end;
						end;
					end;
				else
					self.syncGates = self.syncGates - 1;
				end;
			end;
		end;
	end;
	if self.isServer or g_server ~= nil then
		local usersActive = {};
		for k,v in pairs(g_currentMission.players) do
			table.insert(usersActive, v);
		end;
		if self.userCountActive ~= #usersActive then
			self.userCountActive = #usersActive;
			if g_currentMission.falttorTriggers ~= nil and table.getn(g_currentMission.falttorTriggers) > 0 then
				for i=1, table.getn(g_currentMission.falttorTriggers) do
					if g_currentMission.falttorTriggers[i] ~= nil then
						if g_currentMission.falttorTriggers[i].manualOpen then
							g_server:broadcastEvent(FalttorTriggerManagerEvent:new(g_currentMission.falttorTriggers[i].mpName,g_currentMission.falttorTriggers[i].canBeOpen));
						end;
					end;
				end;
			end;
		end;
	end;
end;
function FalttorTriggerManager:setOpenState(triggerPos, mpName, canBeOpen)
	if triggerPos ~= nil and triggerPos > -1 then
		g_currentMission.falttorTriggers[triggerPos]:setOpenState(canBeOpen);
		if g_server == nil then
			g_client:getServerConnection():sendEvent(FalttorTriggerManagerEvent:new(mpName, canBeOpen));
		else
			g_server:broadcastEvent(FalttorTriggerManagerEvent:new(mpName, canBeOpen));
		end;
	end;
end;
function FalttorTriggerManager:setRandomState(mpName, randomActive, randomStart, randomEnd, randomActiveText)
	g_server:broadcastEvent(FalttorTriggerManagerRandomEvent:new(mpName, randomActive, randomStart, randomEnd, randomActiveText));
end;
function FalttorTriggerManager:draw()
end;
function FalttorTrigger:drawPlayer()
	if self.isClient then
		if self.canDrawOpen == true then
			if self == g_currentMission.player then
				g_currentMission:addHelpButtonText(g_i18n:getText("FALTTOR_OEFFNEN"), InputBinding.OPEN_GATE);
				if g_currentMission.hudTriggerOverlay ~= nil then
					g_currentMission.hudTriggerOverlay:render();
				end;
			end;
		end;
	end;
end;
function FalttorTriggerManager:getFilesCallback(filename)
	if filename == "Falttor_states.xml" then
		self.fileFound = true;
	end;
end;
function FalttorTriggerManager:saveOpenStates()
	local savegame = g_currentMission.missionInfo.savegameDirectory;
	if savegame ~= nil and ((g_server and g_currentMission.missionDynamicInfo.isMultiplayer) or (not g_currentMission.missionDynamicInfo.isMultiplayer)) then
		local xmlFile = createXMLFile("BarrierExtended", self.savegameFilename, "BarrierExtended");
		if xmlFile ~= nil then
			for _,barrierTrigger in pairs(g_currentMission.falttorTriggers) do
				if barrierTrigger.saveName ~= nil then
					local name = "BarrierExtended."..barrierTrigger.saveName;
					setXMLBool(xmlFile, name.."#isValid",barrierTrigger.manualOpen);
					setXMLInt(xmlFile, name.."#canBeOpen",barrierTrigger.canBeOpen);
				end;
			end;
		end;
		saveXMLFile(xmlFile);
		delete(xmlFile);
	end;
end;
function FalttorTriggerManager:loadOpenStates()
	local savegame = g_currentMission.missionInfo.savegameDirectory;
	self.filename = savegame.."/Falttor_states.xml";
	self.fileFound = false;
	getFiles(savegame, "getFilesCallback", self);
	if savegame ~= nil and self.fileFound then
		local xmlFile = loadXMLFile("BarrierExtended", self.savegameFilename);
		if xmlFile ~= nil and g_currentMission.falttorTriggers ~= nil then
			for _,barrierTrigger in pairs(g_currentMission.falttorTriggers) do
				if barrierTrigger.saveName ~= nil then
					local name = "BarrierExtended."..barrierTrigger.saveName;
					local sN = getXMLBool(xmlFile, name.."#isValid");
					if sN ~= nil and sN == true then
						barrierTrigger:setOpenState(Utils.getNoNil(getXMLInt(xmlFile, name.."#canBeOpen"),1));
					end;
				end;
			end;
			delete(xmlFile);
			self.triggersLoaded = true;
		end;
	end;
end;
function FalttorTriggerManager:mouseEvent(posX, posY, isDown, isUp, button)
end;
function FalttorTriggerManager:keyEvent(unicode, sym, modifier, isDown)
end;
function FalttorTriggerRequestEvent:emptyNew()
    local self = Event:new(FalttorTriggerRequestEvent_mt);
    return self;
end;
function FalttorTriggerRequestEvent:new(mpName)
    local self = FalttorTriggerRequestEvent:emptyNew()
    self.mpName = mpName;
    return self;
end;
function FalttorTriggerRequestEvent:readStream(streamId, connection)
    self.mpName = streamReadInt32(streamId);
    self:run(connection);
end;
function FalttorTriggerRequestEvent:writeStream(streamId, connection)
    streamWriteInt32(streamId, self.mpName);
end;
function FalttorTriggerRequestEvent:run(connection)
	if connection:getIsServer() then
		if g_currentMission.falttorTriggers ~= nil and table.getn(g_currentMission.falttorTriggers) > 0 then
			for i=1, table.getn(g_currentMission.falttorTriggers) do
				if g_currentMission.falttorTriggers[i] ~= nil and self.mpName ~= nil and g_currentMission.falttorTriggers[i].mpName == self.mpName then
					g_server:broadcastEvent(FalttorTriggerManagerEvent:new(self.mpName, g_currentMission.falttorTriggers[i].canBeOpen));
				end;
			end;
		end;
	end;
end;
function FalttorTriggerManagerRandomEvent:emptyNew()
    local self = Event:new(FalttorTriggerManagerRandomEvent_mt);
	self.className = "FalttorTrigger";
    return self;
end;
function FalttorTriggerManagerRandomEvent:new(mpName, randomActive, randomStart, randomEnd, randomActiveText)
	local self = FalttorTriggerManagerRandomEvent:emptyNew();
	self.mpName = mpName;
	self.randomActive = randomActive;
	self.randomStart = randomStart;
	self.randomEnd = randomEnd;
	self.randomActiveText = randomActiveText;
	return self;
end;
function FalttorTriggerManagerRandomEvent:writeStream(streamId, connection)
    streamWriteInt32(streamId, self.mpName);
    streamWriteBool(streamId, self.randomActive);
    streamWriteFloat32(streamId, self.randomStart);
    streamWriteFloat32(streamId, self.randomEnd);
    streamWriteString(streamId, self.randomActiveText);
end;
function FalttorTriggerManagerRandomEvent:readStream(streamId, connection)
    self.mpName = streamReadInt32(streamId);
    self.randomActive = streamReadBool(streamId);
    self.randomStart = streamReadFloat32(streamId);
    self.randomEnd = streamReadFloat32(streamId);
    self.randomActiveText = streamReadString(streamId);
    self:run(connection);
end;
function FalttorTriggerManagerRandomEvent:run(connection)
	if g_currentMission.falttorTriggers ~= nil and table.getn(g_currentMission.falttorTriggers) > 0 then
		for i=1, table.getn(g_currentMission.falttorTriggers) do
			if g_currentMission.falttorTriggers[i].mpName == self.mpName then
				g_currentMission.falttorTriggers[i].randomActive = self.randomActive;
				g_currentMission.falttorTriggers[i].randomStart = tonumber(self.randomStart);
				g_currentMission.falttorTriggers[i].randomEnd = tonumber(self.randomEnd);
				g_currentMission.falttorTriggers[i].randomActiveText = self.randomActiveText;
			end;
		end;
	end;
end;
function FalttorTriggerManagerEvent:emptyNew()
    local self = Event:new(FalttorTriggerManagerEvent_mt);
	self.className = "FalttorTrigger";
    return self;
end;
function FalttorTriggerManagerEvent:new(mpName, canBeOpen)
	local self = FalttorTriggerManagerEvent:emptyNew();
	self.mpName = mpName;
	self.canBeOpen = canBeOpen;
	return self;
end;
function FalttorTriggerManagerEvent:writeStream(streamId, connection)
    streamWriteInt32(streamId, self.mpName);
    streamWriteInt8(streamId, self.canBeOpen);
end;
function FalttorTriggerManagerEvent:readStream(streamId, connection)
    self.mpName = streamReadInt32(streamId);
    self.canBeOpen = streamReadInt8(streamId);
    self:run(connection);
end;
function FalttorTriggerManagerEvent:run(connection)
	if g_currentMission.falttorTriggers ~= nil and table.getn(g_currentMission.falttorTriggers) > 0 then
		for i=1, table.getn(g_currentMission.falttorTriggers) do
			if g_currentMission.falttorTriggers[i].mpName == self.mpName then
				g_currentMission.falttorTriggers[i]:setOpenState(self.canBeOpen);
			end;
		end;
	end;
	if not connection:getIsServer() then
		g_server:broadcastEvent(FalttorTriggerManagerEvent:new(self.mpName, self.canBeOpen));
	end;
end;

Steerable.load = Utils.appendedFunction(Steerable.load, FalttorTrigger.load);
Player.load = Utils.appendedFunction(Player.load, FalttorTrigger.load);
Player.draw = Utils.appendedFunction(Player.draw, FalttorTrigger.drawPlayer);
addModEventListener(FalttorTriggerManager);
